#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

// Device inline function for sigmoid
__device__ inline float sigmoidf(float x) {
    return 1.0f / (1.0f + expf(-x));
}

// Kernel: Each warp is assigned to one (batch, channel) tile. Using warp-level primitives to broadcast the per-channel parameters
__global__ void conv3d_warp_kernel(
    const float* __restrict__ output,
    const float* __restrict__ scaling_factor,
    const float* __restrict__ bias,
    float* __restrict__ result,
    const int batch,
    const int channels,
    const int spatial_size) {

    // Identify warp within block
    int warp_in_block = threadIdx.x / 32;
    int lane = threadIdx.x % 32;
    int warps_per_block = blockDim.x / 32;
    
    // Global warp index over (batch, channel) pairs
    int global_warp_id = blockIdx.x * warps_per_block + warp_in_block;

    // Total number of (batch, channel) pairs
    int total_tiles = batch * channels;
    
    if (global_warp_id < total_tiles) {
        // Determine which batch and channel this warp is processing
        int b = global_warp_id / channels;
        int c = global_warp_id % channels;

        // Use lane 0 to load the parameters from global memory and then broadcast them to the entire warp
        float s_val, b_val;
        if (lane == 0) {
            s_val = scaling_factor[c];
            b_val = bias[c];
        }
        s_val = __shfl_sync(0xffffffff, s_val, 0);
        b_val = __shfl_sync(0xffffffff, b_val, 0);

        // Compute the base offset for this (b, c) tile
        int base_idx = (b * channels + c) * spatial_size;

        // Each thread in the warp processes a subset of the spatial elements
        for (int i = lane; i < spatial_size; i += 32) {
            float val = output[base_idx + i];
            val = val * s_val;
            val = tanhf(val);
            val = val * b_val;
            val = sigmoidf(val);
            result[base_idx + i] = val;
        }
    }
}

// Forward function: Performs conv3d then applies scaling, tanh, bias multiplication and sigmoid
// The output tensor is processed in a tiled manner by warps, using warp-level broadcast to fetch per-channel parameters

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor conv_weight,
    torch::Tensor conv_bias,
    torch::Tensor scaling_factor,
    torch::Tensor bias) {

    // Run the built-in conv3d
    auto conv_out = torch::conv3d(x, conv_weight, conv_bias);
    
    int batch = conv_out.size(0);
    int channels = conv_out.size(1);
    int depth = conv_out.size(2);
    int height = conv_out.size(3);
    int width = conv_out.size(4);
    int spatial_size = depth * height * width;

    auto result = torch::empty_like(conv_out);

    // Each warp works on one (batch, channel) tile
    int total_tiles = batch * channels;

    // Set up block dimensions: having, for example, 8 warps per block (8 * 32 = 256 threads)
    int warps_per_block = 8;
    int threads_per_block = warps_per_block * 32; // 32 is 32
    int num_blocks = (total_tiles + warps_per_block - 1) / warps_per_block;

    conv3d_warp_kernel<<<num_blocks, threads_per_block>>>(
        conv_out.data_ptr<float>(),
        scaling_factor.data_ptr<float>(),
        bias.data_ptr<float>(),
        result.data_ptr<float>(),
        batch,
        channels,
        spatial_size
    );

    return result;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Conv3d scale tanh bias sigmoid forward using warp-level broadcast tiling");
}
